/*
* Copyright (c) 2018-2025 (https://github.com/phase1geo/Minder)
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public
* License as published by the Free Software Foundation; either
* version 2 of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* General Public License for more details.
*
* You should have received a copy of the GNU General Public
* License along with this program; if not, write to the
* Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
* Boston, MA 02110-1301 USA
*
* Authored by: Trevor Williams <phase1geo@gmail.com>
*/

@define-color colorPrimary           @GRAPE_700;
@define-color textColorPrimary       @SILVER_100;
@define-color textColorPrimaryShadow @SILVER_500;
@define-color colorAccent            @GRAPE_700;

.theme-selected {
  background: blue;
  color: white;
}

.theme-font-selected {
  color: white;
}

.highlighted {
  background: yellow;
}

.titled {
  font-weight: bold;
}

.shortcut-unselected {
  border: 2px solid transparent;
}

.shortcut-selected {
  border: 2px solid grey;
}

.dim-icon {
  opacity: 0.3;
}

.frameless {
  border: none;
  box-shadow: none;
  background: transparent;
  padding: 0;
}
 
.widget-selected {
  background-color: mix(@theme_base_color, @theme_fg_color, 0.08);
  /* color: @theme_selected_fg_color; */
}
