/*
* Copyright (c) 2018-2025 (https://github.com/phase1geo/Minder)
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public
* License as published by the Free Software Foundation; either
* version 2 of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* General Public License for more details.
*
* You should have received a copy of the GNU General Public
* License along with this program; if not, write to the
* Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
* Boston, MA 02110-1301 USA
*
* Authored by: Trevor Williams <phase1geo@gmail.com>
*/

using Gtk;

public class UndoNodesTagAdd : UndoItem {

  Array<Node> _nodes;
  Tag         _tag;

  //-------------------------------------------------------------
  // Constructor for a node tag add.
  public UndoNodesTagAdd( Array<Node> nodes, Tag tag ) {
    base( _( "tag add" ) );
    _nodes = nodes;
    _tag   = tag;
  }

  //-------------------------------------------------------------
  // Undoes a node tag add.
  public override void undo( MindMap map ) {
    for( int i=0; i<_nodes.length; i++ ) {
      var node = _nodes.index( i );
      node.remove_tag( _tag );
    }
    map.auto_save();
    map.queue_draw();
  }

  //-------------------------------------------------------------
  // Redoes a node tag add.
  public override void redo( MindMap map ) {
    for( int i=0; i<_nodes.length; i++ ) {
      var node = _nodes.index( i );
      node.add_tag( _tag );
    }
    map.auto_save();
    map.queue_draw();
  }

}
